package edu.bu.signstream.common.util;

/**
 * This class converts coordinate to movie time and movie time to coordinate.
 * A coordinate represents a location on a screen, specified in integer precision.
 */

public class CoordinateTimeConverter {

    /**
     * Start movie time of a segment (utterance).
     * Needed for accurate calculation of screen representation of utterance.
     */
    private int segmentStartTime = 0;

    /**
     * Constant, used for calculating movie time representation on a screen.
     * coordinate = pixelInTimeFrame*movieTime
     * By changing pixelInTimeFrame, we can "zoom in" and "zoom out".
     */
    private double pixelInTimeFrame = 0.0124;

	private int timeScale = 600;
	public static int FRAME_SCALE_MPEG = 120000;

	/**
     * Packet duration is a duration of a packet in movie time.
     * It is a difference between movie time of frame x+1 and frame x.
     * Packed duration is a constant for an entire video file.
     */
    public long packetDuration = 0;

	public CoordinateTimeConverter(int segmentStartTime, double pixelInTimeFrame) {
		this.segmentStartTime = segmentStartTime;
		this.pixelInTimeFrame = pixelInTimeFrame;
	}

	public long getPacketDuration() {
		return packetDuration;
	}

	public void setPacketDuration(long packetDuration) {
		this.packetDuration = packetDuration;
	}

	public void setTimeScale(int timeScale) {
		this.timeScale = timeScale;
	}

	public void setPixelInTimeFrame(double pixelInTimeFrame) {
		this.pixelInTimeFrame = pixelInTimeFrame;
	}

	public double getPixelInTimeFrame() {
		return this.pixelInTimeFrame;
	}
	
	public boolean isTimeScaleMPEG() {
		return timeScale == FRAME_SCALE_MPEG;
	}

    /**
     * Calculates preferred screen width for a video segment [duration] long
     * and zoom in index equal to dPixelInTimeFrame
     */
	public int getScreenWidth(int duration, double dPixelInTimeFrame) {
		double a = duration * 30 * 33.33333 * dPixelInTimeFrame;
		double b;
		int abc;
		
		if (isTimeScaleMPEG()) {
			b = a / 200000;
			abc = (int) Math.floor(b);

			// int abc = (int) Math.floor((duration / 200000) * 30 * 33.33333
			// * dPixelInTimeFrame);
			// return abc + 50;
			return abc;
		} else {
			// int abc = (int) Math.floor((duration / 1000) * 30 * 33.33333
			// * dPixelInTimeFrame);
			
			b = a / 1000;
			abc = (int) Math.floor(b);
		}
		
		return abc;
	}

    /**
     * Calculates screen width needed to represent video segment with [duration] long
     * Duration is in movie time.
     */
	public int getScreenWidth(int duration) {
		if (isTimeScaleMPEG()) {
			int abc = (int) Math.floor((duration / 200000) * 30 * 33.33333
					* pixelInTimeFrame);
			return abc + 50;
		} else {
			int abc = (int) Math.floor((duration / 1000) * 30 * 33.33333
					* pixelInTimeFrame);
			return abc + 50;
		}
	}

	/**
     *
     *
     */
    public double calculatePixelInTimeFrame(int packetDuration, int pixels) {
		if (isTimeScaleMPEG()) {
			return 200 * (double) pixels / packetDuration;
		} else {
			return (double) pixels / (double) packetDuration;
		}
	}

	public int getPixelsInOneFrame() {
		if (packetDuration <= 0)
			return 0;

		if (isTimeScaleMPEG()) {
			return (int) (packetDuration * pixelInTimeFrame / 200);
		} else {
			return (int) (packetDuration * pixelInTimeFrame);
		}
	}

	public int convertTimeToCoordinate(int time) {
		int time1 = time;

		if (isTimeScaleMPEG()) {
			int coord = (int) (time1 * pixelInTimeFrame / 200);
			int pixelsInOneFrame = getPixelsInOneFrame();

			if (pixelsInOneFrame >= 2) {
				int frameNumber = (int) (time1 / packetDuration);
				coord = frameNumber * pixelsInOneFrame;
			}
			return coord;
		} else {
			int abstrTime = time1 - segmentStartTime;
			return (int) (abstrTime * pixelInTimeFrame);
		}
	}

	public int convertCoordinateToTime(int coordinate) {
		int cTime = 0;

		if (isTimeScaleMPEG()) {
			int pixelsInOneFrame = getPixelsInOneFrame();
			if (pixelsInOneFrame >= 2) {
				double p = coordinate * packetDuration;
				double r = p / pixelsInOneFrame;
				// System.out.println("r = " + r + ", " + Math.round(r));
				cTime = (int)Math.round(r);
			} else {
				int time = (int) (coordinate / pixelInTimeFrame);
				double t = Math.floor(time / 10);
				cTime = (int) t * 2000;
			}
		} else {
			double abstrTime = coordinate / pixelInTimeFrame;
			int time = (int) abstrTime;
			double t = (time + segmentStartTime) / 10;

			double tt = Math.floor(t);

			double rr = tt / 2;
			double fl = Math.floor(rr);
			if (rr != fl) {
				tt = tt + 1;
			}
			cTime = (int) tt * 10;
		}

		return cTime;
	}

	public int getFrameNumber(int time) {
		return java.lang.StrictMath.round(time / getMovieFrameDuration());
	}

	public int getAdjustedMovieTime(int coord) {
		return getCorrespondingMovieTime(getFrameNumber(convertCoordinateToTime(coord)));
	}

	public int getCorrespondingMovieTime(int frameNumber) {
		return frameNumber * getMovieFrameDuration();
	}

	public int getNextFrameMovieTime(int time) {
		return getCorrespondingMovieTime(getFrameNumber(time) + 1);
	}

	public int getPreviousFrameMovieTime(int time) {
		return getCorrespondingMovieTime(getFrameNumber(time) - 1);
	}

	public int getNextTimeFrameCoordinate(int coord) {
		int time = convertCoordinateToTime(coord);
		int nextTime = getNextFrameMovieTime(time);
		return convertTimeToCoordinate(nextTime);
	}

	public int getMovieFrameDuration() {
		return (int) packetDuration;
	}
}
